import json
import time
import traceback
import threading

import paho.mqtt.client as mqtt

from .sys_utils import IQLog, except_log
from . import settings as S


class MQTTBase(IQLog):
    def __init__(self, client_id="", clean_session=True, logger=None):
        super().__init__(logger)

        self._connected_flag = False

        self._host = "localhost"
        self._port = None

        self._message_callbacks = []
        # self._client = mqtt.Client()
        self._client = mqtt.Client(# mqtt.CallbackAPIVersion.VERSION1,
                                   client_id=client_id,
                                   clean_session=clean_session,
                                   userdata=None,
                                   protocol=mqtt.MQTTv311)

        self._client.on_connect = self._on_connect
        self._client.on_disconnect = self._on_disconnect

    def _connect(self):
        if self._port is not None:
            self._client.connect(self._host, self._port)
        else:
            self._client.connect(self._host)

    @except_log
    def start(self, block=True):
        self.logger.info("Connect to mqtt ({0}:{1})".format(self._host, self._port))
        try:
            self._connect()
        except:
            self.logger.error(traceback.format_exc())
        #     return
        # self.logger.info("Connect to mqtt ({0}:{1}) ... OK".format(self._host, self._port))

        if block:
            self._client.loop_forever()
            self.logger.info("FINISH: mqtt loop")
        else:
            r = self._client.loop_start()
            self.logger.info("START: mqtt loop = {0}".format(r))

    @except_log
    def _on_connect(self, client, userdata, flags, rc):
        txt = "Connection to {0}:{1} returned result: {2}".format(self._host, self._port, mqtt.connack_string(rc))
        if rc != mqtt.CONNACK_ACCEPTED:
            self._connected_flag = False  # Just in case
            self.logger.error(txt)
        else:
            self._connected_flag = True
            self.logger.info(txt)
        if (self._message_callbacks is not None) and self._message_callbacks:
            for topic, callback in self._message_callbacks:
                self._client.message_callback_add(topic, callback)
                self._client.subscribe(topic, 0)

    @except_log
    def _on_disconnect(self, client, userdata, rc):
        self._connected_flag = False
        txt = "Disconnect form {0}:{1} returned result: {2}".format(self._host, self._port, mqtt.error_string(rc))
        if rc != 0:
            self.logger.error("{0} [{1}]".format(txt, "Unexpected disconnection"))
            # self.client.reconnect()  # loop_start()/loop_forever() automatically handles reconnecting
        else:
            self.logger.warning(txt)  # disconnect by disconnect() method of client

    def _publish(self, topic, payload=None, qos=2, retain=True):
        try:
            # self.logger.info("start publish")
            publish_info = self._client.publish(topic, payload, qos, retain)
            # self.logger.info("publish ... ok")
            result = True if publish_info.rc == mqtt.MQTT_ERR_SUCCESS else False
        except:
            result = False
            self.logger.error(traceback.format_exc())
        if not result:
            err_txt = "Cannot send message with topic [{0}]: \"{1}\"".format(topic, payload)
            self.logger.error(err_txt)
        return result

    def _publish_as_json(self, topic, payload=None, qos=2, retain=True):
        try:
            # self.logger.info("get json_payload")
            json_payload = json.dumps(payload, separators=(',', ':'), sort_keys=True)
            # self.logger.info("json_payload ... ok")
        except:
            self.logger.error(traceback.format_exc())
            return False
        return self._publish(topic, json_payload, qos, retain)


def interface_settings_are_changed(settings1, settings2):
    if (settings1 is not None) and (settings2 is not None):
        return settings1 != settings2
    if (settings1 is None) and (settings2 is None):
        return False
    return True


class MQTTClient(MQTTBase):
    def __init__(self, logger=None):
        super().__init__(logger=logger)

        self._message_callbacks = [
            (S.SETTINGS_MQTT_TOPIC, self._on_settings)
        ]

        self._settings_mutex = threading.Lock()
        self._settings = None
        self._new_settings_flag = False

    @except_log
    def _on_settings(self, client, userdata, msg):
        settings = self._process_json_payload(msg)
        # if settings is None:
        #     settings = {}
        self.logger.info("OLD SETTINGS: {0}".format(self._settings))
        self.logger.info("NEW SETTINGS: {0}".format(settings))
        with self._settings_mutex:
            if interface_settings_are_changed(self._settings, settings):
                self._settings = settings
                self._new_settings_flag = True
                self.logger.info("NEW SETTINGS - OK!".format(settings))

    def new_settings(self):
        return self._new_settings_flag

    def get_settings(self):
        with self._settings_mutex:
            self._new_settings_flag = False
            if self._settings is None:
                return None
            return self._settings.copy()

    def _process_json_payload(self, msg):
        try:
            payload = json.loads(msg.payload.decode("utf-8"))
        except:
            payload = None
            self.logger.error(traceback.format_exc())
        return payload

    def publish_result(self, result):
        return self._publish_as_json(S.RESULT_MQTT_TOPIC, result)


class MQTTServerCheck(MQTTBase):
    def __init__(self, mqtt_env, logger=None):
        super().__init__(client_id="{0}_network_manager".format(mqtt_env.NODE_ID),
                         clean_session=True,
                         logger=logger)

        self._host = "{0}.iqnergy.com".format(mqtt_env.NODE_TYPE)
        self._port = int(mqtt_env.PORT)

        self._client.username_pw_set(mqtt_env.LOGIN, mqtt_env.PASSWORD)
        self._client.tls_set(ca_certs=mqtt_env.ROOT_CERT_PATH)

        self.logger.info("{0}:{1} {2}, {3}".format(self._host, self._port, mqtt_env.LOGIN, mqtt_env.ROOT_CERT_PATH))

        self._message_callbacks = [
            (S.SERVER_LASTSEEN_MQTT_TOPIC, self._on_lastseen)
        ]
        self._lastseen = None

    @except_log
    def _on_lastseen(self, client, userdata, msg):
        try:
            lastseen = int(msg.payload.decode("utf-8"))
        except:
            self.logger.error("topic = \"{0}\", payload = \"{1}\"".format(msg.topic, msg.payload))
            self.logger.error(traceback.format_exc())
            return
        self._lastseen = lastseen

    def check(self):
        current_ts = int(time.time())
        self.logger.info("CHECK: curr = {0}, server = {1}".format(current_ts, self._lastseen))
        if self._lastseen is None:
            return False
        if current_ts - self._lastseen < S.SERVER_LASTSEEN_TIMEOUT:
            return True
        return False
